"""Tests upload command of Gencove CLI."""

# pylint: disable=too-many-lines, import-error

import csv
import json
import operator
import os
from uuid import uuid4

from click.testing import CliRunner

from gencove.cli import upload
from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.upload.utils import upload_file
from gencove.constants import ApiEndpoints, UPLOAD_PREFIX
from gencove.models import SampleSheet, UploadSamples, UploadURLImport, UploadsPostData
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import (
    filter_aws_headers,
    filter_jwt,
    replace_gencove_url_vcr,
)
from gencove.tests.upload.vcr.filters import (
    filter_import_fastqs_from_url_response,
    filter_project_samples_request,
    filter_project_samples_response,
    filter_sample_sheet_response,
    filter_upload_credentials_response,
    filter_upload_post_data_response,
    filter_upload_request,
    filter_upload_url_request,
    filter_volatile_dates,
)
from gencove.tests.utils import get_vcr_response
from gencove.utils import get_regular_progress_bar, get_s3_client_refreshable

import pytest  # pylint: disable=wrong-import-order

from vcr import VCR  # pylint: disable=wrong-import-order


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/upload/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
            "X-Amz-Security-Token",
            "X-Amz-Date",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "filter_query_parameters": [("search", "gncv://cli-mock/test.fastq.gz")],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_upload_request,
            filter_upload_url_request,
            filter_project_samples_request,
        ],
        "before_record_response": [
            filter_import_fastqs_from_url_response,
            filter_jwt,
            filter_upload_credentials_response,
            filter_upload_post_data_response,
            filter_volatile_dates,
            filter_aws_headers,
            filter_sample_sheet_response,
            filter_project_samples_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_upload(credentials, vcr, recording, mocker):
    """Sanity check that upload is ok."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")
        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            response = get_vcr_response("/api/v2/uploads-post-data/", vcr)
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**response),
            )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        res = runner.invoke(
            upload,
            ["cli_test_data", *credentials],
        )
        assert not res.exception
        assert res.exit_code == 0
        assert "Uploading cli_test_data/test.fastq.gz to gncv://" in res.output
        assert "All files were successfully uploaded." in res.output
        mocked_get_credentials.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
        mocked_upload_file.assert_called_once()
        assert not mocked_upload_file.call_args[1]["no_progress"]


@pytest.mark.vcr
@assert_authorization
def test_upload_no_files_found(
    credentials, mocker, recording, using_api_key, vcr
):  # pylint: disable=unused-argument
    """Test that no fastq files found exits upload."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            upload,
            ["cli_test_data", *credentials],
        )
        assert res.exit_code == 1
        assert "No FASTQ files found in the path" in res.output
        if using_api_key:
            assert vcr.play_count == 0
        elif not recording:
            # Check that we are making just one request (to the login) if we
            # are not recording the cassette
            assert vcr.play_count == 1


@pytest.mark.vcr
@assert_authorization
def test_upload_invalid_destination(
    credentials, mocker, recording, using_api_key, vcr
):  # pylint: disable=unused-argument
    """Test that invalid destination exists upload."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")
        res = runner.invoke(
            upload,
            ["cli_test_data", "foobar_dir", *credentials],
        )
        assert res.exit_code == 1
        assert (
            f"Invalid destination path. Must start with '{UPLOAD_PREFIX}'" in res.output
        )
        if using_api_key:
            assert vcr.play_count == 0
        elif not recording:
            # Check that we are making just one request (to the login) if we
            # are playing the cassette
            assert vcr.play_count == 1


@pytest.mark.vcr
@assert_authorization
def test_upload_invalid_source(
    credentials, mocker, recording, using_api_key, vcr
):  # pylint: disable=unused-argument
    """Test source raises error if not a map file/directory"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("ACTG")
        res = runner.invoke(
            upload,
            ["cli_test_data/test.fastq.gz", *credentials],
        )
        assert res.exit_code == 1
        assert "must be a directory or a map file" in res.output

        if using_api_key:
            assert vcr.play_count == 0
        elif not recording:
            # Check that we are making just one request (to the login) if we
            # are playing the cassette
            assert vcr.play_count == 1


@pytest.mark.vcr
@assert_authorization
def test_upload_project_id_not_uuid(
    credentials, mocker, recording, using_api_key, vcr
):  # pylint: disable=unused-argument
    """Test that project id is valid UUID when uploading."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")
        res = runner.invoke(
            upload,
            ["cli_test_data", "--run-project-id", "1234", *credentials],
        )

        assert res.exit_code == 1
        assert "--run-project-id is not valid" in res.output
        if using_api_key:
            assert vcr.play_count == 0
        elif not recording:
            # Check that we are making just one request (to the login) if we
            # are playing the cassette
            assert vcr.play_count == 1


@pytest.mark.vcr(
    filter_query_parameters=[("search", "gncv://cli-mock/test.fastq.gz")],
)
@assert_authorization
def test_upload_and_run_immediately(credentials, mocker, project_id, recording, vcr):
    """Upload and assign right away."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )
        res = runner.invoke(
            upload,
            ["cli_test_data", "--run-project-id", project_id, *credentials],
        )

        assert not res.exception
        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately__with_metadata(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                "--run-project-id",
                project_id,
                "--metadata",
                "[1,2]",
                *credentials,
            ],
        )

        assert not res.exception
        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately__invalid_metadata(credentials, mocker, project_id):
    """Upload and assign right away with invalid metadata."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable"
        )

        mocked_get_upload_details = mocker.patch.object(
            APIClient,
            "get_upload_details",
        )
        mocked_upload_file = mocker.patch("gencove.command.upload.main.upload_file")
        mocked_get_sample_sheet = mocker.patch.object(
            APIClient,
            "get_sample_sheet",
        )
        mocked_assign_sample = mocker.patch.object(APIClient, "add_samples_to_project")

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--metadata",
                "[1,2,3",
            ],
        )

        assert res.exit_code == 1
        mocked_get_credentials.assert_not_called()
        mocked_get_upload_details.assert_not_called()
        mocked_upload_file.assert_not_called()
        mocked_get_sample_sheet.assert_not_called()
        mocked_assign_sample.assert_not_called()
        assert "--metadata is not valid JSON" in res.output


@pytest.mark.vcr
@assert_authorization
def test_upload__with_metadata_without_project_id(credentials, mocker):
    """Test that project id is present if attaching metadata when
    uploading."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable"
        )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--metadata",
                "[1,2]",
            ],
        )

        assert res.exit_code == 1
        assert "--metadata cannot be used without --run-project-id" in res.output
        mocked_get_credentials.assert_not_called()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_something_went_wrong(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away did't work."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
        mocked_get_sample_sheet = mocker.patch.object(
            APIClient,
            "get_sample_sheet",
            return_value=SampleSheet(**{"meta": {"next": None}, "results": []}),
        )
        mocked_assign_sample = mocker.patch.object(APIClient, "add_samples_to_project")

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_get_sample_sheet.assert_called()
        mocked_assign_sample.assert_not_called()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_with_output_to_file(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away, then save results to a file."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--output",
                "samples.json",
            ],
        )

        assert not res.exception
        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()

        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
            with open("samples.json", "r", encoding="utf-8") as output_file:
                output_content = output_file.read()
            # r2 fastq key is not present on the file
            assert project_sample_response["uploads"][0]["fastq"]["r2"] is None
            del project_sample_response["uploads"][0]["fastq"]["r2"]
            assert (
                json.dumps(project_sample_response["uploads"], indent=4)
                == output_content
            )


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_with_output_to_nested_file(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away, then save results to a file."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--output",
                "somefolder/samples.json",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()

        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
            with open("somefolder/samples.json", "r", encoding="utf-8") as output_file:
                output_content = output_file.read()
            # r2 fastq key is not present on the file
            assert project_sample_response["uploads"][0]["fastq"]["r2"] is None
            del project_sample_response["uploads"][0]["fastq"]["r2"]
            assert (
                json.dumps(project_sample_response["uploads"], indent=4)
                == output_content
            )


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_with_stdout(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away, then print out the results."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
            # These mocks needs to be here and without side effect otherwise
            # a weird behavior on the stdout will happend and the test fails.
            mocked_upload_file = mocker.patch("gencove.command.upload.main.upload_file")

            mocker.patch(
                "gencove.command.upload.main.get_regular_progress_bar",
            )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--output",
                "-",
            ],
        )

        assert res.exit_code == 0

        mocked_get_credentials.assert_called_once()
        if not recording:
            mocked_upload_file.assert_called_once()
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
            # r2 fastq key is not present on the file
            assert project_sample_response["uploads"][0]["fastq"]["r2"] is None
            del project_sample_response["uploads"][0]["fastq"]["r2"]
            assert (
                json.dumps(project_sample_response["uploads"], indent=4) in res.output
            )


@pytest.mark.vcr
@assert_authorization
def test_upload_without_progressbar(credentials, mocker, recording, vcr):
    """Upload do not show the progress bar."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch("gencove.command.upload.main.upload_file")
        if not recording:
            # Mock get_upload_details only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--no-progress",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        assert mocked_upload_file.call_args[1]["no_progress"]
        if not recording:
            mocked_get_upload_details.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_without_progressbar(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--no-progress",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_not_called()

        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_slow_response_retry(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away and retry on slow response."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
        mocked_get_sample_sheet = mocker.patch.object(
            APIClient,
            "get_sample_sheet",
            side_effect=APIClientTimeout("Could not connect to the api server"),
        )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--no-progress",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
        assert mocked_get_sample_sheet.call_count == 5
        assert "there was an error automatically running them" in res.output


def test_upload_retry_after_unauthorized(mocker):
    """Test that the upload is performed when retrying due to unauthorized
    http response.
    """
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        def _login(
            self, email, password, otp_token=None
        ):  # pylint: disable=unused-argument
            # pylint: disable=protected-access
            self._jwt_refresh_token = "refresh_token"

        mocked_login = mocker.patch.object(
            APIClient, "login", side_effect=_login, autospec=True
        )

        force_refresh_jwt = True

        def _request(
            endpoint,
            *args,  # pylint: disable=unused-argument
            **kwargs,  # pylint: disable=unused-argument
        ):
            if ApiEndpoints.UPLOAD_DETAILS.value == endpoint:
                nonlocal force_refresh_jwt
                if force_refresh_jwt:
                    force_refresh_jwt = False
                    raise APIClientError("Test error.", 401)
                return {
                    "id": str(uuid4()),
                    "last_status": {"id": str(uuid4()), "status": ""},
                    "s3": {"bucket": "test", "object_name": "test"},
                }
            if ApiEndpoints.REFRESH_JWT.value == endpoint:
                return {"access": ""}
            return {}

        mocked_request = mocker.patch.object(
            APIClient, "_request", side_effect=_request
        )

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable"
        )

        res = runner.invoke(
            upload,
            ["cli_test_data"],
            input="\n".join(["foo@bar.com", "123456"]),
        )

        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_get_credentials.assert_called_once()
        # Call count = upload details + refresh jwt + retry upload details
        assert mocked_request.call_count == 3
        assert force_refresh_jwt is False


@pytest.mark.vcr
@assert_authorization
def test_upload_url(credentials, vcr, recording, mocker):
    """Basic check of uploading URL"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        map_file_path = "cli_test_data/test_url_map.fastq-map.csv"
        with open(map_file_path, "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    [
                        "some-id",
                        "r1",
                        "https://s3.amazonaws.com/example/client-id_R1.fastq.gz",
                    ],
                ]
            )
        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )

        if not recording:
            response = get_vcr_response("/api/v2/uploads-url/", vcr)
            mocked_import_fastqs_from_url = mocker.patch.object(
                APIClient,
                "import_fastqs_from_url",
                return_value=UploadURLImport(**response),
            )

        res = runner.invoke(
            upload,
            [map_file_path, *credentials],
        )
        assert not res.exception
        assert res.exit_code == 0
        if not recording:
            mocked_import_fastqs_from_url.assert_called_once()
        mocked_get_credentials.assert_called_once()
        assert "All files were successfully processed" in res.output


def test_upload_url_with_local(credentials):
    """Confirm this case raises error, cannot mix URL and path"""
    # pylint: disable=too-many-locals

    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        map_file_path = "cli_test_data/test_url_map.fastq-map.csv"

        fastq_file_path = "cli_test_data/test.fastq.gz"

        with open(fastq_file_path, "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        with open(map_file_path, "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    [
                        "foo",
                        "r1",
                        "https://s3.amazonaws.com/example/client-id_R1.fastq.gz",
                    ],
                    [
                        "bar",
                        "r1",
                        fastq_file_path,
                    ],
                ]
            )

        res = runner.invoke(
            upload,
            [map_file_path, *credentials],
        )
        assert res.exception
        assert res.exit_code == 1
        assert "Please only supply one type of path" in res.output


@pytest.mark.vcr
@assert_authorization
def test_upload_url_and_run_immediately(
    credentials, vcr, recording, mocker, project_id
):
    """Test uploading then running in project"""
    # pylint: disable=too-many-locals

    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        map_file_path = "cli_test_data/test_url_map.fastq-map.csv"

        with open(map_file_path, "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    [
                        "some-id",
                        "r1",
                        "https://s3.amazonaws.com/example/client-id_R1.fastq.gz",
                    ],
                ]
            )

        if not recording:
            response = get_vcr_response("/api/v2/uploads-url/", vcr)
            mocked_import_fastqs_from_url = mocker.patch.object(
                APIClient,
                "import_fastqs_from_url",
                return_value=UploadURLImport(**response),
            )
            sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )
        res = runner.invoke(
            upload,
            [map_file_path, "--run-project-id", project_id, *credentials],
        )

        assert not res.exception
        assert res.exit_code == 0
        assert "All files were successfully processed" in res.output
        if not recording:
            mocked_import_fastqs_from_url.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()

        mocked_regular_progress_bar.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_default_destination(credentials, vcr, recording, mocker):
    """Test to confirm default destination is gncv://cli-*"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        map_file_path = "cli_test_data/test_map.fastq-map.csv"
        fastq_file_path = "cli_test_data/test.fastq.gz"

        with open(fastq_file_path, "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")

        with open(map_file_path, "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    [
                        "bar",
                        "r1",
                        fastq_file_path,
                    ],
                ]
            )
        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        if not recording:
            response = get_vcr_response("/api/v2/uploads-post-data/", vcr)
            mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**response),
            )
        mocker.patch("gencove.command.upload.main.upload_file", side_effect=upload_file)

        res = runner.invoke(
            upload,
            [map_file_path, *credentials],
        )

    mocked_get_credentials.assert_called_once()

    assert not res.exception
    assert res.exit_code == 0
    assert "uploaded to: gncv://cli-" in res.output
    assert "uploaded to: gncv://cli-url-" not in res.output


@pytest.mark.vcr
@assert_authorization
def test_upload_url_destination(credentials, vcr, recording, mocker):
    """Test to confirm default destination is gncv://cli-url-*"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        map_file_path = "cli_test_data/test_map.fastq-map.csv"

        with open(map_file_path, "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    [
                        "bar",
                        "r1",
                        "https://s3.amazonaws.com/example/client-id_R1.fastq.gz",
                    ],
                ]
            )

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )

        if not recording:
            response = get_vcr_response("/api/v2/uploads-url/", vcr)
            mocker.patch.object(
                APIClient,
                "import_fastqs_from_url",
                return_value=UploadURLImport(**response),
            )

        res = runner.invoke(
            upload,
            [map_file_path, *credentials],
        )

    mocked_get_credentials.assert_called_once()
    assert not res.exception
    assert res.exit_code == 0
    assert "uploaded to: gncv://cli-url-" in res.output



"""Gencove API client.

Exclude imports from linters due to install aliases breaking the rules.
"""
# pylint: disable=too-many-lines

import datetime
import json
import os
import time
from builtins import str as text  # noqa
from urllib.parse import parse_qs, urljoin, urlparse
from uuid import UUID

from pydantic import BaseModel

from requests import (
    ConnectTimeout,
    ReadTimeout,
    delete,
    get,
    post,
)  # noqa: I201

from gencove import constants  # noqa: I100
from gencove.constants import (
    PipelineSortBy,
    SampleArchiveStatus,
    SampleAssignmentStatus,
    SampleSheetSortBy,
    SampleSortBy,
    SampleStatus,
    SortOrder,
)
from gencove.exceptions import MaintenanceError
from gencove.logger import echo_debug
from gencove.models import (  # noqa: I101
    AccessJWT,
    BaseSpaceBiosample,
    BaseSpaceProject,
    BaseSpaceProjectImport,
    BatchDetail,
    CreateJWT,
    FileTypesModel,
    ImportExistingSamplesModel,
    PipelineCapabilities,
    Pipelines,
    PipelineDetail,
    Project,
    ProjectBatches,
    ProjectBatchTypes,
    ProjectMergeVCFs,
    Projects,
    ProjectSamples,
    S3AutoimportTopic,
    S3ProjectImport,
    SampleDetails,
    SampleManifest,
    SampleMetadata,
    SampleQC,
    SampleSheet,
    UploadCredentials,
    UploadSamples,
    UploadURLImport,
    UploadsPostData,
)
from gencove.version import version as cli_version


class CustomEncoder(json.JSONEncoder):
    """JSON encoder that knows how to encode `datetime`, `UUID`
    and `pydantic.BaseModel` objects.
    """

    # pylint: disable=method-hidden
    def default(self, o):
        """Override default method of JSONEncoder."""
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        if isinstance(o, BaseModel):
            return {**o.dict(exclude_unset=True), **o.dict(exclude_none=True)}
        if isinstance(o, UUID):
            return str(o)
        return json.JSONEncoder.default(self, o)


class APIError(Exception):
    """Base API Error."""

    def __init__(self, message, status_code=None):
        super().__init__(message)
        self.message = message
        self.status_code = status_code


class APIClientError(APIError):
    """Base HTTP error."""


class APIClientTooManyRequestsError(APIClientError):
    """Too many requests (429) HTTP error."""

    status_code = 429


class APIClientTimeout(APIClientError):
    """API timeout exception.

    From http://docs.python-requests.org/en/master/user/quickstart/#timeouts

        timeout is not a time limit on the entire response download;
        rather, an exception is raised if the server
        has not issued a response for timeout seconds
        (more precisely, if no bytes have been received on the underlying
        socket for timeout seconds).
        If no timeout is specified explicitly, requests do not time out.

    Catching this error will catch both
    :exc:`~requests.exceptions.ConnectTimeout` and
    :exc:`~requests.exceptions.ReadTimeout` errors.

    ConnectTimeout:
        The request timed out while trying to connect to the remote server.
        Requests that produced this error are safe to retry.

    ReadTimeout:
        The server did not send any data in the allotted amount of time.
    """


# pylint: disable=too-many-public-methods
class APIClient:
    """Gencove API client."""

    endpoints = constants.ApiEndpoints

    def __init__(self, host=None):
        """Initialize api client."""
        self._jwt_token = None
        self._jwt_refresh_token = None
        self._api_key = None
        self.host = host if host is not None else constants.HOST

    @staticmethod
    def _serialize_post_payload(payload):
        return json.dumps(payload, cls=CustomEncoder)

    # pylint: disable=bad-option-value,bad-continuation,too-many-arguments
    # pylint: disable=too-many-branches,too-many-locals, too-many-statements
    def _request(
        self,
        endpoint="",
        params=None,
        method="get",
        custom_headers=None,
        timeout=60,
        sensitive=False,
        raw_response=False,
        files=None,
    ):
        url = urljoin(text(self.host), text(endpoint))
        headers = {
            "content-type": "application/json",
            "date": None,
            "Gencove-cli-version": cli_version(),
        }
        if custom_headers:
            headers.update(custom_headers)

        if not params:
            params = {}

        echo_debug(
            f"Contacting url: {url} with payload: "
            f"{'[SENSITIVE CONTENT]' if sensitive else params}"
        )
        start = time.time()

        try:
            if method == "get":
                response = get(url=url, params=params, headers=headers, timeout=timeout)
            elif method == "delete":
                post_payload = APIClient._serialize_post_payload(params)
                response = delete(
                    url=url,
                    data=post_payload,
                    headers=headers,
                    timeout=timeout,
                )
            else:
                post_payload = APIClient._serialize_post_payload(params)
                if files:
                    # content-type is automatically set by requests library
                    del headers["content-type"]
                    post_payload = None
                response = post(
                    url=url,
                    data=post_payload,
                    headers=headers,
                    timeout=timeout,
                    files=files,
                )

            if response.status_code == 429:
                raise APIClientTooManyRequestsError("Too Many Requests")
        except (ConnectTimeout, ConnectionError):
            # If request timed out,
            # let upper level handle it the way it sees fit.
            # one place might want to retry another might not.
            raise APIClientTimeout(  # pylint: disable=W0707
                "Could not connect to the api server"
            )
        except ReadTimeout:
            raise APIClientTimeout(  # pylint: disable=W0707
                "API server did not respond in timely manner"
            )

        echo_debug(
            f"API response "
            f"is {'[SENSITIVE CONTENT]' if sensitive else response.content} "
            f"status is {response.status_code} in {(time.time() - start) * 1000}ms"
        )

        # pylint: disable=no-member
        if response.status_code >= 200 and response.status_code < 300:
            content = response.text
            if not raw_response:
                content = response.json() if content else {}
            return content

        http_error_msg = ""
        if 400 <= response.status_code < 500:
            http_error_msg = f"API Client Error: {response.reason}"
            if response.text:
                response_json = response.json()
                if "detail" in response_json:
                    http_error_msg += f": {response_json['detail']}"
                else:
                    try:
                        error_msg = "\n".join(
                            [
                                # create-batch can return error details that
                                # is a dict, not a list
                                f"  {key}: {value[0] if isinstance(value, list) else str(value)}"  # noqa: E501  # pylint: disable=line-too-long
                                for key, value in response_json.items()
                            ]
                        )
                    except AttributeError:
                        error_msg = "\n".join(response_json)
                    http_error_msg += f":\n{error_msg}"

        elif 500 <= response.status_code < 600:
            http_error_msg = f"Server Error: {response.reason}"
            if response.status_code == 503:
                if response.text:
                    response_json = response.json()
                    if "maintenance" in response_json:
                        raise MaintenanceError(
                            message=response_json["maintenance_message"],
                            eta=response_json["maintenance_eta"],
                        )

        raise APIClientError(http_error_msg, response.status_code)

    def _set_jwt(self, access_token, refresh_token=None):
        self._jwt_token = access_token
        if refresh_token is not None:
            self._jwt_refresh_token = refresh_token

    def _refresh_authentication(self):
        echo_debug("Refreshing authentication")
        jwt = self.refresh_token(self._jwt_refresh_token)
        self._set_jwt(jwt.access)

    def _get_authorization(self):
        if self._api_key:
            return {"Authorization": f"Api-Key {self._api_key}"}
        return {"Authorization": f"Bearer {self._jwt_token}"}

    def _delete(
        self,
        endpoint,
        payload=None,
        timeout=120,
        authorized=False,
        sensitive=False,
        refreshed=False,
        model=None,
    ):
        headers = {} if not authorized else self._get_authorization()
        try:
            response = self._request(
                endpoint,
                params=payload,
                method="delete",
                timeout=timeout,
                custom_headers=headers,
                sensitive=sensitive,
            )
            if model:
                return model(**response)
            return response
        except APIClientError as err:
            if not refreshed and self._jwt_refresh_token and err.status_code == 401:
                self._refresh_authentication()
                return self._delete(
                    endpoint,
                    payload,
                    timeout,
                    authorized,
                    sensitive,
                    True,
                    model,
                )

            raise err

    def _post(
        self,
        endpoint,
        payload=None,
        timeout=120,
        authorized=False,
        sensitive=False,
        refreshed=False,
        model=None,
        files=None,
    ):
        headers = {} if not authorized else self._get_authorization()
        try:
            response = self._request(
                endpoint,
                params=payload,
                method="post",
                timeout=timeout,
                custom_headers=headers,
                sensitive=sensitive,
                files=files,
            )
            if model:
                return model(**response)
            return response
        except APIClientError as err:
            if not refreshed and self._jwt_refresh_token and err.status_code == 401:
                self._refresh_authentication()
                return self._post(
                    endpoint,
                    payload,
                    timeout,
                    authorized,
                    sensitive,
                    True,
                    model,
                )

            raise err

    def _get(
        self,
        endpoint,
        query_params=None,
        timeout=120,
        authorized=False,
        sensitive=False,
        refreshed=False,
        model=None,
        raw_response=False,
    ):
        headers = {} if not authorized else self._get_authorization()
        try:
            response = self._request(
                endpoint,
                params=query_params,
                method="get",
                timeout=timeout,
                custom_headers=headers,
                sensitive=sensitive,
                raw_response=raw_response,
            )
            if model:
                return model(**response)
            return response
        except APIClientError as err:
            if not refreshed and err.status_code and err.status_code == 401:
                self._refresh_authentication()
                return self._get(
                    endpoint,
                    query_params,
                    timeout,
                    authorized,
                    sensitive,
                    True,
                    model,
                )

            raise err

    @staticmethod
    def _add_query_params(next_link, query_params=None, limit=200):
        if not query_params:
            query_params = {}
        query_params.update({"offset": 0, "limit": limit})
        if next_link:
            query_params["offset"] = parse_qs(urlparse(next_link).query)["offset"]
        return query_params

    def set_api_key(self, api_key):
        """Set api key on this instance."""
        self._api_key = api_key

    def refresh_token(self, refresh_token):
        """Refresh jwt token."""
        return self._post(
            self.endpoints.REFRESH_JWT.value,
            {"refresh": refresh_token},
            sensitive=True,
            model=AccessJWT,
        )

    def get_jwt(self, email, password, otp_token):
        """Get jwt token."""
        body = {"email": email, "password": password}
        if otp_token is not None:
            body["otp_token"] = otp_token
        return self._post(
            self.endpoints.GET_JWT.value,
            body,
            sensitive=True,
            model=CreateJWT,
        )

    def login(self, email, password, otp_token=None):
        """Log user in."""
        jwt = self.get_jwt(email, password, otp_token)
        self._set_jwt(jwt.access, jwt.refresh)

    def get_upload_details(self, gncv_file_path):
        """Get file upload details.
        Args:
            gncv_file_path (str): gncv notated path

        Returns:
            dict: Upload details from the backend
        """
        return self._post(
            self.endpoints.UPLOAD_DETAILS.value,
            dict(destination_path=gncv_file_path),
            authorized=True,
            model=UploadsPostData,
        )

    def get_upload_credentials(self):
        """Get aws credentials for user."""
        return self._post(
            self.endpoints.GET_UPLOAD_CREDENTIALS.value,
            authorized=True,
            sensitive=True,
            model=UploadCredentials,
        )

    def get_project_samples(
        self,
        project_id,
        next_link=None,
        search="",
        sample_status=SampleStatus.ALL.value,
        sample_archive_status=SampleArchiveStatus.ALL.value,
        sort_by=SampleSortBy.MODIFIED.value,
        sort_order=SortOrder.DESC.value,
    ):
        """List single project's associated samples."""
        project_endpoint = self.endpoints.PROJECT_SAMPLES.value.format(id=project_id)
        params = self._add_query_params(
            next_link,
            {
                "search": search,
                "sort_by": sort_by,
                "sort_order": sort_order,
                "status": sample_status,
                "archive_status": sample_archive_status,
            },
        )
        return self._get(
            project_endpoint,
            query_params=params,
            authorized=True,
            model=ProjectSamples,
        )

    def add_samples_to_project(self, samples, project_id, metadata=None):
        """Assign samples to a project.

        Args:
            samples (list of dicts): sample sheet results
            project_id (str): project to which to assign the samples
            metadata (str): Optional JSON metadata to be applied to all samples
        """
        payload = {"uploads": samples, "metadata": metadata}

        return self._post(
            self.endpoints.PROJECT_SAMPLES.value.format(id=project_id),
            payload,
            authorized=True,
            model=UploadSamples,
        )

    def get_sample_details(self, sample_id):
        """Fetch single sample details."""
        return self._get(
            self.endpoints.SAMPLE_DETAILS.value.format(id=sample_id),
            authorized=True,
            model=SampleDetails,
        )

    def get_sample_qc_metrics(self, sample_id):
        """Fetch sample qc metrics.

        Args:
            sample_id(str of uuid): gencove sample id

        Returns:
            list of qc params objects.
        """
        return self._get(
            self.endpoints.SAMPLE_QC_METRICS.value.format(id=sample_id),
            authorized=True,
            model=SampleQC,
            # get max number of items on a single page; a quick fix that avoids
            # paginating through the results for the time being
            query_params={
                "offset": 0,
                "limit": 200,
            },
        )

    def get_sample_sheet(
        self,
        gncv_path=None,
        assigned_status=SampleAssignmentStatus.ALL.value,
        next_link=None,
        sort_by=SampleSheetSortBy.CREATED.value,
        sort_order=SortOrder.DESC.value,
    ):
        """Fetch user samples.

        Args:
            gncv_path (str): filter samples by gncv notated path.
            assigned_status (str, optional, default 'all'): filter samples by
                assignment status. One of SAMPLE_ASSIGNMENT_STATUS.
            next_link (str, optional): url from previous
                response['meta']['next'].
            sort_by(str): upload fastq field to sort by
            sort_order(str): asc or desc sorting
        """
        params = self._add_query_params(
            next_link,
            {
                "search": gncv_path,
                "status": assigned_status,
                "sort_by": sort_by,
                "sort_order": sort_order,
            },
        )
        return self._get(
            self.endpoints.SAMPLE_SHEET.value,
            query_params=params,
            authorized=True,
            model=SampleSheet,
        )

    def list_projects(self, next_link=None):
        """Fetch projects.

        Args:
            next_link (str, optional): url from previous
                response['meta']['next'].

        Returns:
            api response (dict):
                {
                    "meta": {
                        "count": int,
                        "next": str,
                        "previous": optional[str],
                    },
                    "results": [...]
                }
        """
        params = self._add_query_params(next_link)
        return self._get(
            self.endpoints.PROJECTS.value,
            query_params=params,
            authorized=True,
            model=Projects,
        )

    def get_pipeline_capabilities(self, pipeline_id):
        """Get pipeline capabilities details.

        Args:
            pipeline_id (str:uuid): pipeline id

        Returns:
            dict: pipeline details
        """
        resp = self._get(
            self.endpoints.PIPELINE_CAPABILITES.value.format(id=pipeline_id),
            authorized=True,
            model=PipelineCapabilities,
        )
        return resp

    def get_pipeline_capabilities_for_pipeline(self, pipeline_id):
        """Get pipeline capabilities details.

        Args:
            pipeline_id (str:uuid): pipeline id

        Returns:
            dict: pipeline details
        """
        resp = self._get(
            self.endpoints.PIPELINE.value.format(id=pipeline_id),
            authorized=True,
            model=PipelineDetail,
        )
        return resp

    def get_project_batch_types(self, project_id, next_link=None):
        """List single project's available batch types."""
        project_endpoint = self.endpoints.PROJECT_BATCH_TYPES.value.format(
            id=project_id
        )
        params = self._add_query_params(next_link)
        return self._get(
            project_endpoint,
            query_params=params,
            authorized=True,
            model=ProjectBatchTypes,
        )

    def get_project_batches(self, project_id, next_link=None):
        """List single project's batches."""
        project_endpoint = self.endpoints.PROJECT_BATCHES.value.format(id=project_id)
        params = self._add_query_params(next_link)
        return self._get(
            project_endpoint,
            query_params=params,
            authorized=True,
            model=ProjectBatches,
        )

    def create_project_batch(self, project_id, batch_type, batch_name, sample_ids):
        """Making a post request to create project batch."""
        project_endpoint = self.endpoints.PROJECT_BATCHES.value.format(id=project_id)

        payload = {
            "name": batch_name,
            "batch_type": batch_type,
            "sample_ids": sample_ids,
        }

        return self._post(
            project_endpoint, payload, authorized=True, model=ProjectBatches
        )

    def get_batch(self, batch_id):
        """Get single batch."""
        batches_endpoint = self.endpoints.BATCHES.value.format(id=batch_id)
        return self._get(batches_endpoint, authorized=True, model=BatchDetail)

    def create_sample_manifest(self, project_id, sample_manifest):
        """Make a post request to create project sample manifest."""
        project_endpoint = self.endpoints.PROJECT_SAMPLE_MANIFESTS.value.format(
            id=project_id
        )

        with open(sample_manifest, "rb") as manifest_rb:
            return self._post(
                project_endpoint,
                authorized=True,
                files={
                    "sample_manifest": (os.path.basename(sample_manifest), manifest_rb)
                },
            )

    def get_sample_manifest(self, manifest_id):
        """Make a get request to get sample manifest."""
        project_endpoint = self.endpoints.SAMPLE_MANIFESTS.value.format(
            id=manifest_id,
        )
        return self._get(project_endpoint, authorized=True, model=SampleManifest)

    def get_sample_manifests(self, project_id):
        """Make a get request to get project sample manifests."""
        project_endpoint = self.endpoints.PROJECT_SAMPLE_MANIFESTS.value.format(
            id=project_id
        )
        return self._get(project_endpoint, authorized=True)

    def restore_project_samples(self, project_id, sample_ids):
        """Make a request to restore samples in given project."""
        restore_project_samples_endpoint = (
            self.endpoints.PROJECT_RESTORE_SAMPLES.value.format(id=project_id)
        )

        payload = {"sample_ids": sample_ids}

        return self._post(restore_project_samples_endpoint, payload, authorized=True)

    def get_project(self, project_id):
        """Get single project."""
        project_endpoint = f"{self.endpoints.PROJECTS.value}{project_id}"
        return self._get(project_endpoint, authorized=True, model=Project)

    def create_merged_vcf(self, project_id):
        """Merge VCF files for a project."""
        merge_vcf_endpoint = self.endpoints.PROJECT_MERGE_VCFS.value.format(
            id=project_id
        )
        return self._post(merge_vcf_endpoint, authorized=True, model=ProjectMergeVCFs)

    def retrieve_merged_vcf(self, project_id):
        """Retrieve the status of the merge command for a project."""
        merge_vcf_endpoint = self.endpoints.PROJECT_MERGE_VCFS.value.format(
            id=project_id
        )
        return self._get(merge_vcf_endpoint, authorized=True, model=ProjectMergeVCFs)

    def get_metadata(self, sample_id):
        """Retrieve the metadata for a sample."""
        sample_metadata_endpoint = self.endpoints.SAMPLE_METADATA.value.format(
            id=sample_id
        )
        return self._get(
            sample_metadata_endpoint, authorized=True, model=SampleMetadata
        )

    def set_metadata(self, sample_id, metadata):
        """Assign the metadata to a sample."""
        sample_metadata_endpoint = self.endpoints.SAMPLE_METADATA.value.format(
            id=sample_id
        )
        payload = {
            "metadata": metadata,
        }
        return self._post(
            sample_metadata_endpoint,
            payload,
            authorized=True,
            model=SampleMetadata,
        )

    def import_basespace_projects(
        self, basespace_project_ids, project_id, metadata=None
    ):
        """Make a request to import Biosamples from BaseSpace projects to
        a given project.

        Args:
            basespace_project_ids (list of strings): BaseSpace projects
            project_id (str): project to which to assign the samples
            metadata (str): Optional JSON metadata to be applied to all samples
        """

        payload = {
            "basespace_project_ids": basespace_project_ids,
            "project_id": project_id,
            "metadata": metadata,
        }

        return self._post(
            self.endpoints.BASESPACE_PROJECTS_IMPORT.value,
            payload,
            authorized=True,
        )

    def list_basespace_projects(self, next_link=None):
        """Make a request to list BaseSpace projects.

        Args:
            next_link (str, optional): url from previous
                response['meta']['next'].

        Returns:
            api response (dict):
                {
                    "meta": {
                        "count": int,
                        "next": str,
                        "previous": optional[str],
                    },
                    "results": [...]
                }
        """
        params = self._add_query_params(next_link)
        return self._get(
            self.endpoints.BASESPACE_PROJECTS_LIST.value,
            query_params=params,
            authorized=True,
            model=BaseSpaceProject,
        )

    def list_biosamples(self, basespace_project_id, next_link=None):
        """Make a request to list Biosamples from a BaseSpace project.

        Args:
            basespace_project_id (str): what BaseSpace project to request
                Biosamples from
            next_link (str, optional): url from previous
                response['meta']['next'].

        Returns:
            api response (dict):
                {
                    "meta": {
                        "count": int,
                        "next": str,
                        "previous": optional[str],
                    },
                    "results": [...]
                }
        """
        params = self._add_query_params(next_link)
        return self._get(
            self.endpoints.BASESPACE_BIOSAMPLES_LIST.value.format(
                id=basespace_project_id
            ),
            query_params=params,
            authorized=True,
            model=BaseSpaceBiosample,
        )

    def s3_uri_import(self, s3_uri, project_id, metadata=None):
        """Make a request to import samples from S3 URI to a given
        project.

        Args:
            s3_uri (str): s3 path formated as s3://<bucket-name>/prefix
            project_id (str): project to which to assign the samples
            metadata (str): Optional JSON metadata to be applied to all samples
        """

        payload = {
            "s3_uri": s3_uri,
            "project_id": project_id,
            "metadata": metadata,
        }

        return self._post(
            self.endpoints.S3_URI_IMPORT.value,
            payload,
            authorized=True,
        )

    def autoimport_from_basespace(
        self,
        project_id,
        identifier,
        metadata=None,
    ):
        """Make a request to create a periodic import job of BaseSpace
        projects' Biosamples to a given Gencove project.

        Args:
            project_id (str): project to which to assign the samples
            identifier (str): identifier that is contained in BaseSpace
                projects' name
            metadata (str): Optional JSON metadata to be applied to all samples
        """
        payload = {
            "project_id": project_id,
            "identifier": identifier,
            "metadata": metadata,
        }

        return self._post(
            self.endpoints.BASESPACE_PROJECTS_AUTOIMPORT.value,
            payload,
            authorized=True,
        )

    def list_basespace_autoimport_jobs(self, next_link=None):
        """Make a request to list periodic import jobs of BaseSpace projects'
        Biosamples.

        Args:
            next_link (str, optional): url from previous
                response['meta']['next'].

        Returns:
            api response (dict):
                {
                    "meta": {
                        "count": int,
                        "next": str,
                        "previous": optional[str],
                    },
                    "results": [...]
                }
        """

        params = self._add_query_params(next_link)
        return self._get(
            self.endpoints.BASESPACE_PROJECTS_AUTOIMPORT.value,
            query_params=params,
            authorized=True,
            model=BaseSpaceProjectImport,
        )

    def autoimport_from_s3(
        self,
        project_id,
        s3_uri,
        metadata=None,
    ):
        """Make a request to create an import job of S3 URI to a given
        Gencove project.

        Args:
            s3_uri (str): s3 path formated as s3://<bucket-name>/prefix
            project_id (str): project to which to assign the samples
            metadata (str): Optional JSON metadata to be applied to all samples
        """

        payload = {
            "project_id": project_id,
            "s3_uri": s3_uri,
            "metadata": metadata,
        }

        return self._post(
            self.endpoints.S3_URI_AUTOIMPORT.value,
            payload,
            authorized=True,
            model=S3AutoimportTopic,
        )

    def list_s3_autoimport_jobs(self, next_link=None):
        """Make a request to list S3 import jobs.

        Args:
            next_link (str, optional): url from previous
                response['meta']['next'].

        Returns:
            api response (dict):
                {
                    "meta": {
                        "count": int,
                        "next": str,
                        "previous": optional[str],
                    },
                    "results": [...]
                }
        """

        params = self._add_query_params(next_link)
        return self._get(
            self.endpoints.S3_URI_AUTOIMPORT.value,
            query_params=params,
            authorized=True,
            model=S3ProjectImport,
        )

    def delete_project_samples(self, project_id, sample_ids):
        """Make a request to delete samples in given project.

        Args:
            project_id (str): project to which to assign the samples
            sample_ids (list of strings): sample_ids to soft delete
        """
        delete_project_samples_endpoint = (
            self.endpoints.PROJECT_DELETE_SAMPLES.value.format(id=project_id)
        )

        payload = {"sample_ids": sample_ids}

        return self._delete(delete_project_samples_endpoint, payload, authorized=True)

    def delete_projects(self, project_ids):
        """Make a request to delete projects.

        Args:
            project_ids (List[str]): projects to delete
        """
        delete_projects_endpoint = self.endpoints.PROJECTS_DELETE.value

        payload = {"project_ids": project_ids}

        return self._delete(delete_projects_endpoint, payload, authorized=True)

    def get_file_checksum(self, file_id, filename=None):
        """Fetch file checksum, using the client because need to be
        authenticated.
        """
        params = {}
        if filename:
            params["filename"] = filename
        return self._get(
            self.endpoints.FILE_CHECKSUM.value.format(id=file_id),
            query_params=params,
            authorized=True,
            raw_response=True,
        )

    def import_existing_samples(self, project_id, sample_ids, metadata):
        """Import existing samples to a project and pass metadata."""
        payload = {
            "project_id": project_id,
            "samples": [{"sample_id": sample_id} for sample_id in sample_ids],
        }
        if metadata:
            payload["metadata"] = metadata
        return self._post(
            self.endpoints.IMPORT_EXISTING_SAMPLES.value,
            payload,
            authorized=True,
            model=ImportExistingSamplesModel,
        )

    def import_fastqs_from_url(self, gncv_file_path, url):
        """POST fastq URL data to API"""
        payload = {
            "destination_path": gncv_file_path,
            "source_url": url,
        }
        return self._post(
            self.endpoints.UPLOAD_URL.value,
            payload,
            authorized=True,
            model=UploadURLImport,
        )

    def get_file_types(self, project_id=None):
        """List file types.

        Args:
            project_id (UUID, optional): project_id used to get all file types
                for a project
        """
        params = {}
        if project_id:
            params["project_id"] = project_id
        return self._get(
            self.endpoints.FILE_TYPES.value,
            query_params=params,
            authorized=True,
            model=FileTypesModel,
        )

    def get_pipelines(self, next_link=None):
        """List pipelines.

        Args:
            next_link (str, optional): url from previous
                response['meta']['next'].

        Returns:
            api response (dict):
                {
                    "meta": {
                        "count": int,
                        "next": str,
                        "previous": optional[str],
                    },
                    "results": [...]
                }
        """
        params = self._add_query_params(
            next_link,
            {
                "sort_by": PipelineSortBy.CREATED.value,
                "sort_order": SortOrder.DESC.value,
            },
        )
        return self._get(
            self.endpoints.PIPELINES.value,
            query_params=params,
            authorized=True,
            model=Pipelines,
        )

    def create_project(self, project_name, pipeline_capability_id):
        """Making a post request to create a project."""
        project_endpoint = self.endpoints.PROJECTS.value

        payload = {
            "name": project_name,
            "pipeline_capabilities": pipeline_capability_id,
        }

        return self._post(project_endpoint, payload, authorized=True, model=Project)



"""Test available pipelines list command."""
# pylint: disable=wrong-import-order, import-error
import operator

from click.testing import CliRunner


from gencove.client import APIClient, APIClientTimeout
from gencove.command.projects.cli import list_project_pipelines
from gencove.models import Pipelines
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_project_pipelines_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_project_pipelines_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_list_pipelines__empty(credentials, mocker, recording, vcr):
    """Test there are no pipelines."""
    runner = CliRunner()
    if not recording:
        get_pipelines_response = get_vcr_response(
            "/api/v2/pipeline/", vcr, operator.contains
        )
        mocked_get_pipelines = mocker.patch.object(
            APIClient,
            "get_pipelines",
            return_value=Pipelines(**get_pipelines_response),
        )
    res = runner.invoke(
        list_project_pipelines,
        [*credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_pipelines.assert_called_once()
    assert res.output == ""


@pytest.mark.vcr
@assert_authorization
def test_list_pipelines__not_empty(mocker, credentials, recording, vcr):
    """Test pipelines being output to the shell."""
    runner = CliRunner()
    if not recording:
        # Mock get_pipelines only if using the cassettes, since we
        # mock the return value.
        get_pipelines_response = get_vcr_response(
            "/api/v2/pipeline/", vcr, operator.contains
        )
        get_pipelines_response["meta"]["count"] = 200
        get_pipelines_response["meta"]["next"] = None
        mocked_get_pipelines = mocker.patch.object(
            APIClient,
            "get_pipelines",
            return_value=Pipelines(**get_pipelines_response),
        )
    res = runner.invoke(
        list_project_pipelines,
        [*credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_pipelines.assert_called_once()
        pipelines = "\n".join(
            [
                f"{pipeline['id']}\t{pipeline['version']}"
                for pipeline in get_pipelines_response["results"]
            ]
        )
        assert res.output.startswith(pipelines) is True


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_pipelines__not_empty_slow_response_retry(credentials, mocker):
    """Test pipelines slow response retry."""
    runner = CliRunner()
    mocked_get_pipelines = mocker.patch.object(
        APIClient,
        "get_pipelines",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    res = runner.invoke(
        list_project_pipelines,
        [*credentials],
    )
    assert res.exit_code == 1
    assert mocked_get_pipelines.call_count == 2



"""Test project list command."""
# pylint: disable=wrong-import-order, import-error
import io
import operator
import os
import sys
from datetime import datetime, timedelta
from platform import platform
from uuid import uuid4

import boto3

from click import echo
from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
    APIClientTimeout,
)  # noqa: I100
from gencove.command.projects.cli import list_projects
from gencove.models import (
    PipelineCapabilities,
    Project,
    Projects,
)
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_list_projects_response,
    filter_pipeline_capabilities_request,
    filter_pipeline_capabilities_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response
from gencove.version import version

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_pipeline_capabilities_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_list_projects_response,
            filter_pipeline_capabilities_response,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_empty(mocker, credentials):
    """Test user organization has no projects."""
    runner = CliRunner()
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "list_projects",
        return_value=Projects(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(list_projects, credentials)
    assert res.exit_code == 0
    mocked_get_projects.assert_called_once()
    assert "" in res.output


MOCKED_PIPELINE_CAPABILITY = {
    "id": str(uuid4()),
    "name": "test capability",
    "private": False,
    "merge_vcfs_enabled": False,
}


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_projects_no_permission(mocker, credentials):
    """Test projects no permission available to show them."""
    runner = CliRunner()
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "list_projects",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=403
        ),
        return_value={"detail": "Not found"},
    )
    mocked_get_pipeline_capabilities = mocker.patch.object(
        APIClient, "get_pipeline_capabilities"
    )
    res = runner.invoke(list_projects, credentials)
    assert res.exit_code == 1
    mocked_get_projects.assert_called_once()
    mocked_get_pipeline_capabilities.assert_not_called()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\n".join(
            [
                "ERROR: You do not have the sufficient permission "
                "level required to perform this operation.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_projects_slow_response_retry_list(mocker, credentials):
    """Test projects slow response retry on the list."""
    runner = CliRunner()
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "list_projects",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )
    mocked_get_pipeline_capabilities = mocker.patch.object(
        APIClient, "get_pipeline_capabilities"
    )
    res = runner.invoke(list_projects, credentials)
    assert res.exit_code == 1
    assert mocked_get_projects.call_count == 2
    mocked_get_pipeline_capabilities.assert_not_called()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_projects_slow_response_dump_log(mocker, credentials, dump_filename):
    """Test projects slow response dumps a log filet."""
    del os.environ["GENCOVE_SAVE_DUMP_LOG"]
    runner = CliRunner()
    mocker.patch.object(
        APIClient,
        "list_projects",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )
    python_version = sys.version_info
    logs = [
        f"Python version: {python_version.major}.{python_version.minor}.{python_version.micro}",  # noqa: E501  # pylint: disable=line-too-long
        f"CLI version: {version()}",
        f"OS details: {platform()}",
        f"boto3 version: {boto3.__version__}",
        "Retrieving projects",
        "Get projects page",
        "Could not connect to the api server",
    ]
    with runner.isolated_filesystem():
        res = runner.invoke(list_projects, credentials)
        assert res.exit_code == 1
        assert (
            f"Please attach the debug log file located in {dump_filename} to a bug report."  # noqa: E501  # pylint: disable=line-too-long
            in res.output
        )
        with open(dump_filename, encoding="utf8") as log_file:
            log_content = log_file.read()
            for log in logs:
                assert log in log_content


@pytest.mark.vcr
@assert_authorization
def test_list_projects_slow_response_retry_pipeline(
    mocker, credentials, recording, vcr
):
    """Test projects slow repsonse retry on the pipeline capabilities."""
    runner = CliRunner()
    if not recording:
        # Mock list_projects only if using the cassettes, since we mock the
        # return value.
        list_projects_response = get_vcr_response("/api/v2/projects/", vcr)
        mocked_get_projects = mocker.patch.object(
            APIClient,
            "list_projects",
            return_value=Projects(**list_projects_response),
        )
    mocked_get_pipeline_capabilities = mocker.patch.object(
        APIClient,
        "get_pipeline_capabilities",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )
    res = runner.invoke(list_projects, credentials)
    assert res.exit_code == 1
    if not recording:
        mocked_get_projects.assert_called_once()
    assert mocked_get_pipeline_capabilities.call_count == 3


@pytest.mark.vcr
@assert_authorization
def test_list_projects(mocker, credentials, recording, vcr):
    """Test projects being outputed to the shell."""
    runner = CliRunner()
    if not recording:
        # Mock list_projects only if using the cassettes, since we mock the
        # return value.
        list_projects_response = get_vcr_response("/api/v2/projects/", vcr)
        mocked_get_projects = mocker.patch.object(
            APIClient,
            "list_projects",
            return_value=Projects(**list_projects_response),
        )
        get_pipeline_capabilities_response = get_vcr_response(
            "/api/v2/pipeline-capabilities/", vcr, operator.contains
        )
        mocked_get_pipeline_capabilities = mocker.patch.object(
            APIClient,
            "get_pipeline_capabilities",
            return_value=PipelineCapabilities(**get_pipeline_capabilities_response),
        )
    res = runner.invoke(list_projects, credentials)
    assert res.exit_code == 0
    if not recording:
        mocked_get_projects.assert_called_once()
        projects = list_projects_response["results"]
        assert mocked_get_pipeline_capabilities.call_count == len(projects)
        output_line = io.BytesIO()
        sys.stdout = output_line
        for project in projects:
            project = Project(**project)
            echo(
                "\t".join(
                    [
                        str(project.created),
                        str(project.id),
                        project.name.replace("\t", " "),
                        get_pipeline_capabilities_response["name"],
                    ]
                )
            )
        assert output_line.getvalue() == res.output.encode()


@pytest.mark.vcr
@assert_authorization
def test_list_projects_with_capabilities(mocker, credentials, recording, vcr):
    """Test projects being outputed to the shell."""
    runner = CliRunner()
    if not recording:
        # Mock list_projects only if using the cassettes, since we mock the
        # return value.
        list_projects_response = get_vcr_response("/api/v2/projects/", vcr)
        mocked_get_projects = mocker.patch.object(
            APIClient,
            "list_projects",
            return_value=Projects(**list_projects_response),
        )
        get_pipeline_capabilities_response = get_vcr_response(
            "/api/v2/pipeline-capabilities/", vcr, operator.contains
        )
        mocked_get_pipeline_capabilities = mocker.patch.object(
            APIClient,
            "get_pipeline_capabilities",
            return_value=PipelineCapabilities(**get_pipeline_capabilities_response),
        )
    res = runner.invoke(list_projects, ["--include-capability", *credentials])
    assert res.exit_code == 0
    if not recording:
        mocked_get_projects.assert_called_once()
        projects = list_projects_response["results"]
        assert mocked_get_pipeline_capabilities.call_count == len(projects)
        assert "key" in get_pipeline_capabilities_response
        projects = res.output.splitlines()
        for project in projects:
            # Confirm # of columns is correct
            assert len(project.split("\t")) == 6


# API responses may return new keys and values eventually
MOCKED_PROJECTS_WITH_UNEXPECTED_KEYS = dict(
    meta=dict(next=None),
    results=[
        {
            "id": str(uuid4()),
            "name": "test\tproject",
            "description": "",
            "created": (datetime.utcnow() - timedelta(days=7)).isoformat(),
            "organization": str(uuid4()),
            "webhook_url": "",
            "sample_count": 1,
            "pipeline_capabilities": str(uuid4()),
            "roles": [],
            **{"unexpected_key" + str(uuid4()): i for i in range(10)},
        }
    ],
)


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_projects__with_unexpected_keys(mocker, credentials):
    """Test projects being outputed to the shell where webhook_url, roles and
    some randomly generated values are part of the response.
    """

    runner = CliRunner()
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "list_projects",
        return_value=Projects(**MOCKED_PROJECTS_WITH_UNEXPECTED_KEYS),
    )
    mocked_get_pipeline_capabilities = mocker.patch.object(
        APIClient,
        "get_pipeline_capabilities",
        return_value=PipelineCapabilities(**MOCKED_PIPELINE_CAPABILITY),
    )
    res = runner.invoke(list_projects, credentials)
    assert res.exit_code == 0
    mocked_get_projects.assert_called_once()
    mocked_get_pipeline_capabilities.assert_called_once()

    project = Project(**MOCKED_PROJECTS_WITH_UNEXPECTED_KEYS["results"][0])
    pipeline = PipelineCapabilities(**MOCKED_PIPELINE_CAPABILITY)

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                str(project.created),
                str(project.id),
                project.name.replace("\t", " "),
                pipeline.name,
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Define common exceptions in Gencove CLI."""


class ValidationError(Exception):
    """CLI validation error - indicating command input is not valid."""

    def __init__(self, message=""):
        super().__init__(message)
        self.message = message


class MaintenanceError(Exception):
    """API maintenance error - indicating backend API is in maintenance mode."""

    def __init__(self, message="", eta=""):
        super().__init__(message)
        self.message = message
        self.eta = eta



"""Utilities for processing imported samples."""


def get_line(imported_sample):
    """Build a list of relevant data to be printed.

    Args:
        imported_sample (SampleImport): instance of SampleImport

    Returns:
        str: relevant data to be printed
    """

    return "\t".join(
        [
            str(imported_sample.sample_id),
            imported_sample.client_id if imported_sample.client_id else "",
        ]
    )



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .biosamples_list.cli import biosamples_list


@click.group()
def biosamples():
    """BaseSpace Biosamples management commands."""


biosamples.add_command(biosamples_list)



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .autoimports.cli import autoimports
from .biosamples.cli import biosamples
from .projects.cli import projects


@click.group()
def basespace():
    """BaseSpace managements commands."""


basespace.add_command(autoimports)
basespace.add_command(biosamples)
basespace.add_command(projects)



"""Test project's delete samples command."""
# pylint: disable=wrong-import-order, import-error
import operator
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.projects.cli import delete_project_samples
from gencove.exceptions import MaintenanceError
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_project_delete_samples,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_project_delete_samples,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
        ],
    }


@assert_no_requests
def test_delete_project_samples__bad_project_id(
    mocker, credentials
):  # pylint: disable=unused-argument
    """Test delete project samples when non-uuid string is used as project
    id."""
    runner = CliRunner()

    res = runner.invoke(
        delete_project_samples,
        [
            "22222222",
            *credentials,
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,"
            "22222222-2222-2222-2222-222222222222",
        ],
    )
    assert res.exit_code == 1
    assert "Project ID is not valid" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_delete_project_samples__not_owned_project(credentials, mocker):
    """Test delete project samples failure when project is not owned."""
    runner = CliRunner()
    mocked_delete_project_samples = mocker.patch.object(
        APIClient,
        "delete_project_samples",
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        delete_project_samples,
        [
            str(uuid4()),
            *credentials,
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,"
            "22222222-2222-2222-2222-222222222222",
        ],
    )
    assert res.exit_code == 1
    mocked_delete_project_samples.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


@pytest.mark.vcr
@assert_authorization
def test_delete_project_samples__success__empty_sample_ids(
    credentials, mocker, project_id, recording, vcr
):
    """Test delete project samples success when an empty list of sample ids
    is sent."""

    runner = CliRunner()
    if not recording:
        # Mock delete_project_samples only if using the cassettes, since we
        # mock the return value.
        delete_project_samples_response = get_vcr_response(
            "/api/v2/project-delete-samples/", vcr, operator.contains
        )
        mocked_delete_project_samples = mocker.patch.object(
            APIClient,
            "delete_project_samples",
            return_value=delete_project_samples_response,
        )

    res = runner.invoke(
        delete_project_samples,
        [
            project_id,
            *credentials,
            "--sample-ids",
            "",
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_delete_project_samples.assert_called_once()
    assert "The following samples have been deleted successfully" in res.output


@assert_no_requests
def test_delete_project_samples__invalid_sample_ids(
    credentials, mocker
):  # pylint: disable=unused-argument
    """Test delete project samples failure when an invalid list of sample ids
    is sent."""

    runner = CliRunner()

    res = runner.invoke(
        delete_project_samples,
        [
            str(uuid4()),
            *credentials,
            "--sample-ids",
            "1111,222",
        ],
    )
    assert res.exit_code == 1
    assert "Not all sample IDs are valid" in res.output


@pytest.mark.vcr
@assert_authorization
def test_delete_project_samples__sample_not_in_project(
    credentials, mocker, project_id, recording, vcr
):
    """Test delete project samples with sample not in project."""

    runner = CliRunner()
    if not recording:
        # Mock delete_project_samples only if using the cassettes, since we
        # mock the return value.
        delete_project_samples_response = get_vcr_response(
            "/api/v2/project-delete-samples/",
            vcr,
            operator.contains,
            just_body=False,
        )
        mocked_delete_project_samples = mocker.patch.object(
            APIClient,
            "delete_project_samples",
            side_effect=APIClientError(
                message=delete_project_samples_response["body"]["string"],
                status_code=delete_project_samples_response["status"]["code"],
            ),
        )
    res = runner.invoke(
        delete_project_samples,
        [
            project_id,
            *credentials,
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111",
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_delete_project_samples.assert_called_once()
    assert "All sample ids must be part of the current project." in res.output


@pytest.mark.vcr
@assert_authorization
@pytest.mark.default_cassette("test_delete_project_samples__success.yaml")
@pytest.mark.parametrize("remove_hyphens", [True, False])
def test_delete_project_samples__success(  # pylint: disable=too-many-arguments
    deleted_sample, credentials, mocker, project_id, recording, vcr, remove_hyphens
):
    """Test delete project samples success."""
    runner = CliRunner()
    if not recording:
        # Mock delete_project_samples only if using the cassettes, since we
        # mock the return value.
        delete_project_samples_response = get_vcr_response(
            "/api/v2/project-delete-samples/", vcr, operator.contains
        )
        mocked_delete_project_samples = mocker.patch.object(
            APIClient,
            "delete_project_samples",
            return_value=delete_project_samples_response,
        )

    if remove_hyphens:
        project_id = project_id.replace("-", "")

    res = runner.invoke(
        delete_project_samples,
        [
            project_id,
            *credentials,
            "--sample-ids",
            deleted_sample,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_delete_project_samples.assert_called_once()
    assert "The following samples have been deleted successfully" in res.output


@pytest.mark.vcr
@assert_authorization
@pytest.mark.default_cassette(
    "test_delete_project_samples__returns_maintenance_error_503.yaml"
)
def test_delete_project_samples__returns_maintenance_error_503(
    # pylint: disable=too-many-arguments
    deleted_sample,
    credentials,
    mocker,
    project_id,
    recording,
    using_api_key,
    vcr,
):
    """Test delete project samples success."""
    runner = CliRunner()
    if not recording:
        # Mock delete_project_samples only if using the cassettes, since we
        # mock the return value.

        delete_project_samples_response = get_vcr_response(
            "/api/v2/project-delete-samples/", vcr, operator.contains
        )
        mocked_delete_project_samples = mocker.patch.object(
            APIClient,
            "delete_project_samples",
            side_effect=MaintenanceError(
                message=delete_project_samples_response["maintenance_message"],
                eta=delete_project_samples_response["maintenance_eta"],
            ),
        )

    res = runner.invoke(
        delete_project_samples,
        [
            project_id,
            *credentials,
            "--sample-ids",
            deleted_sample,
        ],
    )
    assert res.exit_code == 1
    if not recording:
        if using_api_key:
            mocked_delete_project_samples.assert_called_once()
        else:
            mocked_delete_project_samples.assert_not_called()

    assert (
        "ERROR: Gencove is currently undergoing maintenance and "
        "will return at the given ETA. "
        "Thank you for your patience."
    ) in res.output



"""Test project's get merged VCF command."""

from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.projects.cli import get_merged_vcf
from gencove.models import Project


def test_get_merged_vcf__bad_project_id(mocker):
    """Test get merged file failure when non-uuid string is used as
    project id.
    """
    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        return_value=Project(id=str(uuid4())),
    )

    res = runner.invoke(
        get_merged_vcf,
        [
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_not_called()
    mocked_get_project.assert_not_called()
    assert "Project ID is not valid" in res.output


def test_get_merged_vcf__not_owned_project(mocker):
    """Test get merged file failure when project is not owned."""
    mocked_response = {"detail": "Not found."}

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        return_value=mocked_response,
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        get_merged_vcf,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_project.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


def test_get_merged_vcf__empty(mocker):
    """Test project doesn't have a merged VCF file."""
    project_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        return_value=Project(
            id=project_id,
            name="Project Cadmus",
            description="",
            created="2020-06-11T02:14:00.541889Z",
            organization=str(uuid4()),
            sample_count=3,
            pipeline_capabilities=str(uuid4()),
            files=[],
        ),
    )

    res = runner.invoke(
        get_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_project.assert_called_once()
    assert f"No files to process for project {project_id}" in res.output


def test_get_merged_vcf_custom_filename(mocker):
    """Test project download merged VCF success with custom filename."""
    project_id = str(uuid4())
    file_id = str(uuid4())
    download_url = (
        "https://bucket.s3.amazonaws.com/output/apps/merge_vcfs/"
        f"{file_id}/{file_id}.vcf.bgz"
    )

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        return_value=Project(
            id=project_id,
            name="Project Cadmus",
            description="",
            created="2020-06-11T02:14:00.541889Z",
            organization=str(uuid4()),
            sample_count=3,
            pipeline_capabilities=str(uuid4()),
            files=[
                {
                    "id": "755ec682-e4a5-414a-a5be-07e0af11cf75",
                    "s3_path": (
                        "app-data/output/apps/merge_vcfs/"
                        f"{file_id}/{file_id}.vcf.bgz"
                    ),
                    "size": None,
                    "download_url": download_url,
                    "file_type": "impute-vcf-merged",
                }
            ],
        ),
    )

    with runner.isolated_filesystem():
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_merged_vcf.main.download.utils."
            "download_file"
        )
        res = runner.invoke(
            get_merged_vcf,
            [
                project_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--output-filename",
                "superman.vcf.gz",
            ],
        )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project.assert_called_once()
    mocked_download_file.assert_called_once_with(
        "superman.vcf.gz", download_url, no_progress=False
    )


def test_get_merged_vcf__no_progress_success(mocker):
    """Test project download merged VCF success."""
    project_id = str(uuid4())
    file_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        return_value=Project(
            id=project_id,
            name="Project Cadmus",
            description="",
            created="2020-06-11T02:14:00.541889Z",
            organization=str(uuid4()),
            sample_count=3,
            pipeline_capabilities=str(uuid4()),
            files=[
                {
                    "id": "755ec682-e4a5-414a-a5be-07e0af11cf75",
                    "s3_path": (
                        "app-data/output/apps/merge_vcfs/"
                        f"{file_id}/{file_id}.vcf.bgz"
                    ),
                    "size": None,
                    "download_url": (
                        "https://bucket.s3.amazonaws.com/output/apps/"
                        f"merge_vcfs/{file_id}/{file_id}.vcf.bgz"
                    ),
                    "file_type": "impute-vcf-merged",
                }
            ],
        ),
    )

    with runner.isolated_filesystem():
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_merged_vcf.main.download.utils."
            "download_file"
        )
        res = runner.invoke(
            get_merged_vcf,
            [
                project_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--no-progress",
            ],
        )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project.assert_called_once()
    mocked_download_file.assert_called_once_with(
        f"{file_id}.vcf.bgz",
        "https://bucket.s3.amazonaws.com/output/apps/"
        f"merge_vcfs/{file_id}/{file_id}.vcf.bgz",
        no_progress=True,
    )


def test_get_merged_vcf__slow_response_retry(mocker):
    """Test project download merged VCF slow response retry."""
    project_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    with runner.isolated_filesystem():
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_merged_vcf.main.download.utils."
            "download_file"
        )
        res = runner.invoke(
            get_merged_vcf,
            [
                project_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    assert mocked_get_project.call_count == 5
    mocked_download_file.assert_not_called()


def test_get_merged_vcf__success(mocker):
    """Test project download merged VCF success."""
    project_id = str(uuid4())
    file_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        return_value=Project(
            id=project_id,
            name="Project Cadmus",
            description="",
            created="2020-06-11T02:14:00.541889Z",
            organization=str(uuid4()),
            sample_count=3,
            pipeline_capabilities=str(uuid4()),
            files=[
                {
                    "id": "755ec682-e4a5-414a-a5be-07e0af11cf75",
                    "s3_path": (
                        "app-data/output/apps/merge_vcfs/"
                        f"{file_id}/{file_id}.vcf.bgz"
                    ),
                    "size": None,
                    "download_url": (
                        "https://bucket.s3.amazonaws.com/output/apps/"
                        f"merge_vcfs/{file_id}/{file_id}.vcf.bgz"
                    ),
                    "file_type": "impute-vcf-merged",
                }
            ],
        ),
    )

    with runner.isolated_filesystem():
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_merged_vcf.main.download.utils."
            "download_file"
        )
        res = runner.invoke(
            get_merged_vcf,
            [
                project_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project.assert_called_once()
    mocked_download_file.assert_called_once_with(
        f"{file_id}.vcf.bgz",
        "https://bucket.s3.amazonaws.com/output/apps/"
        f"merge_vcfs/{file_id}/{file_id}.vcf.bgz",
        no_progress=False,
    )


def test_get_merged_vcf__success__project_with_legacy_webhhok_url(mocker):
    """Test project download merged VCF success."""
    project_id = str(uuid4())
    file_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project = mocker.patch.object(
        APIClient,
        "get_project",
        return_value=Project(
            id=project_id,
            name="Project Cadmus",
            description="",
            created="2020-06-11T02:14:00.541889Z",
            organization=str(uuid4()),
            webhook_url="",
            sample_count=3,
            pipeline_capabilities=str(uuid4()),
            files=[
                {
                    "id": "755ec682-e4a5-414a-a5be-07e0af11cf75",
                    "s3_path": (
                        "app-data/output/apps/merge_vcfs/"
                        f"{file_id}/{file_id}.vcf.bgz"
                    ),
                    "size": None,
                    "download_url": (
                        "https://bucket.s3.amazonaws.com/output/apps/"
                        f"merge_vcfs/{file_id}/{file_id}.vcf.bgz"
                    ),
                    "file_type": "impute-vcf-merged",
                }
            ],
        ),
    )

    with runner.isolated_filesystem():
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_merged_vcf.main.download.utils."
            "download_file"
        )
        res = runner.invoke(
            get_merged_vcf,
            [
                project_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project.assert_called_once()
    mocked_download_file.assert_called_once_with(
        f"{file_id}.vcf.bgz",
        "https://bucket.s3.amazonaws.com/output/apps/"
        f"merge_vcfs/{file_id}/{file_id}.vcf.bgz",
        no_progress=False,
    )



"""Create project's batch executor."""
import os

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.exceptions import ValidationError


class CreateSampleManifest(Command):
    """Create project's sample manifest executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        sample_manifest,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.sample_manifest = sample_manifest

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if self.sample_manifest[-4:] != ".csv":
            raise ValidationError(
                "sample_manifest argument must be a local path to a CSV file"
            )

        if not os.path.exists(self.sample_manifest):
            raise ValidationError(
                "supplied path to sample_manifest argument does not exist"
            )

    def execute(self):
        """Make a request to create a sample manifest for given project."""
        self.echo_debug(f"Creating sample manifest for project {self.project_id} ")

        try:
            created_sample_manifest_details = self.api_client.create_sample_manifest(
                project_id=self.project_id, sample_manifest=self.sample_manifest
            )
            self.echo_debug(created_sample_manifest_details)
            self.echo_info(
                f"Successfully uploaded sample manifest to project {self.project_id}"
            )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error creating the sample manifest.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(f"Project {self.project_id} does not exist.")
            else:
                raise



"""Python library which enables you to use Gencoves' research backend."""
import click

from gencove import version
from gencove.command.basespace import basespace
from gencove.command.download import download
from gencove.command.files import list_file_types
from gencove.command.projects import projects
from gencove.command.s3_imports import s3
from gencove.command.sample_manifests import sample_manifests
from gencove.command.samples import samples
from gencove.command.upload import upload
from gencove.command.uploads import uploads
from gencove.command.webhook import webhooks
from gencove.utils import python_version_check


def announcements():
    """Preamble announcements displayed whenever the CLI is called"""
    python_version_check()


@click.group()
@click.version_option(version=version.version())
def cli():
    """Gencove's command line interface."""


announcements()
cli.add_command(basespace)
cli.add_command(download)
cli.add_command(list_file_types)
cli.add_command(upload)
cli.add_command(uploads)
cli.add_command(projects)
cli.add_command(samples)
cli.add_command(sample_manifests)
cli.add_command(s3)
cli.add_command(webhooks)

if __name__ == "__main__":
    cli()



"""Constants for import-existing-samples command."""
from typing import Optional

from gencove.constants import Optionals


# pylint: disable=too-few-public-methods
class ImportExistingSamplesOptionals(Optionals):
    """ImportExistingSamplesOptionals model"""

    metadata_json: Optional[str]



"""Download command utilities."""
import json
import os
import re
from urllib.parse import parse_qs, urlparse

import backoff

import requests

from gencove import client  # noqa: I100
from gencove.constants import (  # noqa: I100
    DownloadTemplateParts,
    MAX_RETRY_TIME_SECONDS,
)
from gencove.logger import echo_debug, echo_info, echo_warning
from gencove.models import SampleFile
from gencove.utils import get_progress_bar

from .constants import (
    CHUNK_SIZE,
    DEFAULT_FILENAME_TOKEN,
    FILENAME_RE,
    FILE_TYPES_MAPPER,
    FilePrefix,
)


def _get_prefix_parts(full_prefix):
    """Extract directories prefix parts from."""
    prefix_parts = full_prefix.split("/")
    file_name, _, file_ext = prefix_parts[-1].partition(".")
    return FilePrefix(
        dirs="/".join(prefix_parts[:-1]),
        filename=file_name,
        file_extension=file_ext,
        use_default_filename=(DEFAULT_FILENAME_TOKEN in full_prefix),
    )


def get_filename_from_download_url(url):
    """Deduce filename from url.

    Args:
        url (str): URL string

    Returns:
        str: filename
    """
    try:
        filename = re.findall(
            FILENAME_RE,
            parse_qs(urlparse(url).query)["response-content-disposition"][0],
        )[0]
    except (KeyError, IndexError):
        echo_debug(
            "URL didn't contain filename query argument. Assume filename from url"
        )
        filename = urlparse(url).path.split("/")[-1]

    return filename


def deliverable_type_from_filename(filename):
    """Deduce deliverable type based on dot notation."""
    filetype = ".".join(filename.split(".")[1:])
    echo_debug(f"Deduced filetype to be: {filetype} from filename: {filename}")
    return filetype


def _create_filepath(download_to, prefix_dirs, filename):
    """Build full file path and ensure that directory structure exists.

    Args:
        download_to (str): top level directory path
        prefix_dirs (str): subdirectories structure to create under
            download_to.
        filename (str): name of the file inside download_to/file_prefix
            structure.
    """
    echo_debug(f"_create_filepath Downloading to: {download_to}")
    echo_debug(f"_create_filepath file prefix is: {prefix_dirs}")

    path = os.path.join(download_to, prefix_dirs)
    # Cross-platform cross-python-version directory creation
    if not os.path.exists(path):
        echo_debug(f"creating path: {path}")
        os.makedirs(path)

    file_path = os.path.join(path, filename)
    echo_debug(f"Deduced full file path is {file_path}")
    return file_path


def build_file_path(deliverable, file_with_prefix, download_to, filename=None):
    """Create and return file system path where the file will be downloaded to.

    Args:
        deliverable (dict): used to get download url and file type
        file_with_prefix (str): used as a template for download path
        download_to (str): general location where to download the file to

    Returns:
         str : file path on current file system
    """
    prefix = _get_prefix_parts(file_with_prefix)
    download_url, file_type = "", ""
    if isinstance(deliverable, SampleFile):
        download_url, file_type = (
            deliverable.download_url,
            deliverable.file_type,
        )
    else:
        download_url, file_type = (
            deliverable.get("download_url"),
            deliverable.get("file_type"),
        )
    source_filename = (
        filename if filename else get_filename_from_download_url(download_url)
    )

    destination_filename = prefix.filename
    if prefix.file_extension:
        destination_filename = f"{prefix.filename}.{prefix.file_extension}"

    # turning off formatting for improved code readability
    # fmt: off
    destination_filename = destination_filename.format(
        **{
            DownloadTemplateParts.FILE_TYPE.value: FILE_TYPES_MAPPER.get(file_type) or file_type,  # noqa: E501  # pylint: disable=line-too-long
            DownloadTemplateParts.FILE_EXTENSION.value: deliverable_type_from_filename(source_filename),  # noqa: E501  # pylint: disable=line-too-long
            DownloadTemplateParts.DEFAULT_FILENAME.value: source_filename,
        }
    )
    # fmt: on

    echo_debug(f"Calculated destination filename: {destination_filename}")
    return _create_filepath(download_to, prefix.dirs, destination_filename)


def fatal_request_error(err=None):
    """Give up retrying if the error code is in fatal range.

    Returns:
        bool: True if to giveup on backing off, False it to continue.
    """
    if not err or not err.response:
        return False
    if err.response.status_code == 403:
        # download url needs to be refreshed, give up on backoff
        return True
    # retry 4xx or 5xx and all else not
    return not 400 <= err.response.status_code <= 600


@backoff.on_exception(
    backoff.expo,
    (
        requests.exceptions.HTTPError,
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
    ),
    max_time=MAX_RETRY_TIME_SECONDS,
    giveup=fatal_request_error,
)
def download_file(file_path, download_url, skip_existing=True, no_progress=False):
    """Download a file to file system.

    Args:
        file_path (str): full file path, according to destination
            and download template
        download_url (str): url of the file to download
        skip_existing (bool): skip already downloaded files
        no_progress (bool): don't show progress bar

    Returns:
        str : file path
            location of the downloaded file
    """

    file_path_tmp = f"{file_path}.tmp"
    if os.path.exists(file_path_tmp):
        file_mode = "ab"
        headers = dict(Range=f"bytes={os.path.getsize(file_path_tmp)}-")
        echo_info(f"Resuming previous download: {file_path}")
    else:
        file_mode = "wb"
        headers = {}
        echo_info(f"Downloading file to {file_path}")

    stream_params = dict(
        stream=True, allow_redirects=False, headers=headers, timeout=30
    )

    with requests.get(download_url, **stream_params) as req:
        req.raise_for_status()
        total = int(req.headers["content-length"])
        # pylint: disable=E0012,C0330
        if (
            skip_existing
            and os.path.isfile(file_path)
            and os.path.getsize(file_path) == total
        ):
            echo_info(f"Skipping existing file: {file_path}")
            return file_path

        echo_debug(f"Starting to download file to: {file_path}")

        with open(file_path_tmp, file_mode) as downloaded_file:
            if not no_progress:
                pbar = get_progress_bar(
                    int(req.headers["content-length"]), "Downloading: "
                )
                pbar.start()
            for chunk in req.iter_content(chunk_size=CHUNK_SIZE):
                downloaded_file.write(chunk)
                if not no_progress:
                    pbar.update(pbar.value + len(chunk))
            if not no_progress:
                pbar.finish()

        # Cross-platform cross-python-version file overwriting
        if os.path.exists(file_path):
            echo_debug(f"Found old file under same name: {file_path}. Removing it.")
            os.remove(file_path)
        os.rename(file_path_tmp, file_path)
        echo_info(f"Finished downloading a file: {file_path}")
        return file_path


def save_metadata_file(path, api_client, sample_id, skip_existing=True):
    """Helper function to save metadata to json file.

    Args:
        path(str): full file path where metadata will be saved
        api_client(APIClient): an instance of the client to make a call
        sample_id(str of uuid): sample gencove id
        skip_existing(bool): option of skipping the existing file

    Returns:
        None
    """
    if skip_existing and os.path.isfile(path) and os.path.getsize(path):
        echo_info(f"Skipping existing file: {path}")
        return
    try:
        metadata = api_client.get_metadata(sample_id)
    except client.APIClientError:
        echo_warning("Error getting sample metadata.")
        raise
    echo_info(f"Downloading file to: {path}")
    with open(path, "w", encoding="utf-8") as metadata_file:
        metadata_file.write(metadata.json())
    echo_info(f"Finished downloading a file: {path}")


def save_qc_file(path, api_client, sample_id, skip_existing=True):
    """Helper function to save qc metrics to json file.

    Args:
        path(str): full file path where qc metrics will be saved
        api_client(APIClient): an instance of the client to make a call
        sample_id(str of uuid): sample gencove id
        skip_existing(bool): option of skipping the existing file

    Returns:
        None
    """
    if skip_existing and os.path.isfile(path) and os.path.getsize(path):
        echo_info(f"Skipping existing file: {path}")
        return
    try:
        sample_qcs = api_client.get_sample_qc_metrics(sample_id).results
    except client.APIClientError:
        echo_warning("Error getting sample quality control metrics.")
        raise
    echo_info(f"Downloading file to: {path}")
    with open(path, "w", encoding="utf-8") as qc_file:
        content = json.dumps(sample_qcs, cls=client.CustomEncoder)
        qc_file.write(content)
    echo_info(f"Finished downloading a file: {path}")


def fatal_process_sample_error(err):
    """Give up retrying if the error code is different from 403.

    If the error code is 403, we need to refresh the download url and try
    processing the sample again.

    Returns:
        bool: True if to giveup on backing off, False it to continue.
    """
    return err.response.status_code != 403


def get_download_template_format_params(client_id, gencove_id):
    """Return format parts for download template.

    Args:
        client_id (str): sample client id
        gencove_id (str): sample gencove id

    Returns:
        format parts : dict
    """
    return {
        DownloadTemplateParts.CLIENT_ID.value: client_id,
        DownloadTemplateParts.GENCOVE_ID.value: gencove_id,
        DownloadTemplateParts.FILE_TYPE.value: f"{{{DownloadTemplateParts.FILE_TYPE.value}}}",  # noqa: E501  # pylint: disable=line-too-long
        DownloadTemplateParts.FILE_EXTENSION.value: f"{{{DownloadTemplateParts.FILE_EXTENSION.value}}}",  # noqa: E501  # pylint: disable=line-too-long
        DownloadTemplateParts.DEFAULT_FILENAME.value: f"{{{DownloadTemplateParts.DEFAULT_FILENAME.value}}}",  # noqa: E501  # pylint: disable=line-too-long
    }



"""Webhook signature verification shell command definition."""
import click

from gencove.logger import echo_error, echo_info

from .utils import is_valid_signature


@click.command("verify")
@click.argument("secret")
@click.argument("header")
@click.argument("payload")
# pylint: disable=too-many-arguments
def verify(
    secret,
    header,
    payload,
):
    """Verify webhook signature.

    `SECRET`: key to be used as a secret for hmac algorithm.

    `HEADER`: Gencove-Signature header content.

    `PAYLOAD`: JSON payload (i.e., the request’s body).
    """
    if is_valid_signature(secret, header, payload):
        echo_info("Webhook payload successfully verified.")
    else:
        echo_error("Could not verify webhook payload.")
        raise click.Abort()



"""Utilities for listing merged VCF files."""


def get_line(merged_vcf):
    """Build a list of relevant data to be printed.

    Args:
        merged_vcf (dict): an object from project merge VCF response

    Returns:
        str: tab delimited list of relevant data to be printed
    """
    return "\t".join(
        [
            str(merged_vcf.id),
            merged_vcf.last_status.created.isoformat(),
            merged_vcf.last_status.status,
        ]
    )



"""Sample set metadata shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import SetMetadata


@click.command("set-metadata")
@click.argument("sample_id")
@click.option(
    "--json",
    help="JSON string of the metadata to be assigned to a sample.",
    required=True,
)
@add_options(common_options)
# pylint: disable=too-many-arguments
def set_metadata(
    sample_id,
    json,
    host,
    email,
    password,
    api_key,
):
    """Set sample metadata.

    `SAMPLE_ID`: sample for which to set the metadata
    """
    SetMetadata(
        sample_id,
        json,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""VCR filters for download tests."""
from gencove.tests.filters import _replace_uuid_from_url
from gencove.tests.utils import MOCK_CHECKSUM, MOCK_UUID


def filter_project_samples_request(request):
    """Filter project samples sensitive data from request."""
    return _replace_uuid_from_url(request, "project-samples")


def filter_files_request(request):
    """Filter file sensitive data from request."""
    return _replace_uuid_from_url(request, "files")


def filter_files_response(response):
    """Filter file sensitive data from response."""
    if ".sha256" in response["headers"].get("Content-Disposition", [""])[0]:
        filename = f"{MOCK_UUID}_R2.fastq.gz.sha256"
        response["headers"]["Content-Disposition"] = [
            f'attachment; filename="{filename}"'
        ]
        response["body"][
            "string"
        ] = f"{MOCK_CHECKSUM} *{MOCK_UUID}_R2.fastq.gz\n".encode()
    return response



"""List pipeline capabilities subcommand."""

import backoff

from gencove import client
from gencove.command.base import Command

from .utils import get_line


class ListPipelineCapabilities(Command):
    """List pipeline capabilities command executor."""

    def __init__(self, pipeline_id, credentials, options):
        super().__init__(credentials, options)
        self.pipeline_id = pipeline_id

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    def execute(self):
        self.echo_debug("Retrieving possible pipeline capabilities:")

        try:
            pipeline_capabilities = self.get_pipeline_capabilities()
            for pipeline_capability in pipeline_capabilities.capabilities:
                self.echo_data(get_line(pipeline_capability))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error listing pipeline capabilities.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_error(f"Pipeline {self.pipeline_id} does not exist.")
            raise

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_pipeline_capabilities(self):
        """Get pipeline capabilities page."""
        return self.api_client.get_pipeline_capabilities_for_pipeline(
            pipeline_id=self.pipeline_id,
        )



"""Import existing samples to a project subcommand."""
import json

from .utils import get_line
from ...base import Command
from ...utils import is_valid_json
from .... import client
from ....exceptions import ValidationError


class ImportExistingSamples(Command):
    """Import existing samples command executor."""

    def __init__(self, project_id, sample_ids, credentials, options):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.sample_ids = sample_ids
        self.metadata_json = options.metadata_json

    def initialize(self):
        """Initialize import-existing-samples subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if self.metadata_json and is_valid_json(self.metadata_json) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")

    def execute(self):
        """Import existing samples to the given project."""
        metadata = None
        if self.metadata_json is not None:
            metadata = json.loads(self.metadata_json)
            self.echo_info("Assigning metadata to the importing samples.")
        self.echo_info(f"Import existing samples to the project: {self.project_id}")
        try:
            # Import existing samples to the project optionally
            # passing metadata.
            import_existing_samples_response = self.api_client.import_existing_samples(
                self.project_id, self.sample_ids, metadata
            )
            for imported_sample in import_existing_samples_response.samples:
                self.echo_data(get_line(imported_sample))
            self.echo_info(
                f"Number of samples imported into the project {self.project_id}: "
                f"{len(self.sample_ids)}"
            )
            if metadata:
                self.echo_info("Metadata attached to each sample.")
        except client.APIClientError as err:
            self.echo_debug(err.message)
            self.echo_error("There was an error importing samples.")
            raise



"""Samples list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, SampleAssignmentStatus
from gencove.utils import enum_as_dict

from .constants import UploadsOptions
from .main import ListSampleSheet


@click.command("list")
@click.option("--search", help="Filter uploads by gncv path")
@click.option(
    "--status",
    help="Filter uploads by status of assignment",
    type=click.Choice(enum_as_dict(SampleAssignmentStatus).values()),
    default=SampleAssignmentStatus.ALL.value,
)
@add_options(common_options)
def list_uploads(  # pylint: disable=E0012,C0330,R0913
    search, status, host, email, password, api_key
):
    """List user uploads."""
    ListSampleSheet(
        Credentials(email=email, password=password, api_key=api_key),
        UploadsOptions(host=host, status=status, search=search),
    ).run()



"""Download sample file subcommand."""

import backoff

import requests

from gencove.command.download.constants import (  # noqa: I100
    ALLOWED_ARCHIVE_STATUSES_RE,
    METADATA_FILE_TYPE,
    QC_FILE_TYPE,
)

from .utils import (
    download_file,
    fatal_process_sample_error,
    save_metadata_file,
    save_qc_file,
)
from ...base import Command
from ...utils import is_valid_uuid
from .... import client
from ....exceptions import ValidationError


class DownloadFile(Command):
    """Download file command executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        sample_id,
        file_type,
        destination,
        credentials,
        options,
        no_progress,
        checksum,
    ):
        super().__init__(credentials, options)
        self.sample_id = sample_id
        self.file_type = file_type
        self.destination = destination
        self.no_progress = no_progress
        self.checksum = checksum

    def initialize(self):
        """Initialize download-file subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

        if is_valid_uuid(self.sample_id) is False:
            raise ValidationError("Sample ID is not valid. Exiting.")

    def execute(self):
        self.echo_debug("Retrieving sample file:")

        try:
            self.get_file()
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 404:
                self.echo_error(
                    f"Sample {self.sample_id} file {self.file_type} does not "
                    "exist or you do not have permission required to access it."
                )
            raise

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.HTTPError,
        giveup=fatal_process_sample_error,
        max_tries=10,
    )
    def get_file(self):
        """Retrieves file and writes it to destination"""
        try:
            sample = self.api_client.get_sample_details(self.sample_id)
        except client.APIClientError:
            self.echo_warning(
                f"Sample with id {self.sample_id} not found. "
                "Are you using client id instead of sample id?"
            )
            return

        self.echo_debug(
            f"Processing sample id {sample.id}, status {sample.last_status.status}"
        )

        if not ALLOWED_ARCHIVE_STATUSES_RE.match(sample.archive_last_status.status):
            raise ValidationError(
                f"Sample with id {sample.id} is archived and cannot be downloaded - "
                "please restore the sample and try again."
            )

        sample_file_types = {file.file_type for file in sample.files}.union(
            {QC_FILE_TYPE, METADATA_FILE_TYPE}
        )
        if self.file_type not in sample_file_types:
            raise ValidationError(
                f"Sample with id {sample.id} does not have any files with "
                f"file type {self.file_type}. Valid file types for this sample are: "
                f"{', '.join(sorted(sample_file_types))}."
            )

        file_to_download = None

        for sample_file in sample.files:
            # pylint: disable=E0012,C0330
            if self.file_type == sample_file.file_type:
                file_to_download = sample_file
                break

        if self.file_type == QC_FILE_TYPE:
            save_qc_file(self.destination, self.api_client, self.sample_id)
        elif self.file_type == METADATA_FILE_TYPE:
            save_metadata_file(self.destination, self.api_client, self.sample_id)
        elif file_to_download is None or file_to_download.download_url is None:
            self.echo_warning(
                f"File not found for sample with id {self.sample_id} "
                f"and file type {self.file_type}"
            )
        else:
            download_file(
                self.destination,
                file_to_download.download_url,
                self.no_progress,
            )
            writing_to_stdout = self.destination.isatty()
            if not writing_to_stdout and self.checksum:
                try:
                    checksum = self.api_client.get_file_checksum(
                        file_to_download.id,
                        filename=self.destination.name,
                    )
                    self.create_checksum_file(self.destination.name, checksum)
                except client.APIClientTooManyRequestsError:
                    self.echo_debug(
                        "Request was throttled for checksum file, trying again"
                    )
                    raise

    def create_checksum_file(self, file_path, checksum_sha256):
        """Create checksum file.

        Args:
            file_path (str): File path of the original file,
                will append .sha256
            checksum_sha256 (str): Checksum (sha256) value for the file

        Returns:
            None
        """
        checksum_path = f"{file_path}.sha256"
        self.echo_debug(f"Adding checksum file: {checksum_path}")
        with open(checksum_path, "w", encoding="utf-8") as checksum_file:
            checksum_file.write(checksum_sha256)



"""Get sample manifest executor."""
from pathlib import Path

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.download.utils import download_file


class GetSampleManifest(Command):
    """Get sample manifest executor."""

    def __init__(
        self,
        manifest_id,
        destination,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.manifest_id = manifest_id
        self.destination = destination

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

    def execute(self):
        """Request to get sample manifest and download it."""
        self.echo_debug(f"Retrieving sample manifest with ID {self.manifest_id}")

        try:
            sample_manifest = self.api_client.get_sample_manifest(
                manifest_id=self.manifest_id,
            )
            dst = (
                Path(self.destination)
                / str(sample_manifest.id)
                / sample_manifest.file_name
            )
            dst.parent.mkdir(exist_ok=True, parents=True)
            self.echo_info(
                f"Downloading manifest with ID {sample_manifest.id} to {dst}"
            )
            download_file(
                file_path=str(dst),
                download_url=sample_manifest.file.download_url,
                no_progress=True,
            )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error retrieving the sample manifest.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(
                    f"Sample manifest with ID {self.manifest_id} does not exist or "
                    "you do not have access."
                )
            else:
                raise



"""Merge project's VCF files executor."""
from ..status_merged_vcf.utils import get_line
from ...base import Command
from .... import client


class CreateMergedVCF(Command):
    """Merge project's VCF files executor."""

    def __init__(
        self,
        project_id,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id

    def initialize(self):
        """Initialize subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to merge VCF files for a given project."""
        self.echo_debug(f"Merging VCF files for project {self.project_id}")
        try:
            created_merge_details = self.api_client.create_merged_vcf(self.project_id)
            self.echo_debug(created_merge_details)
            self.echo_info(f"Issued merge request for project {self.project_id}")
            self.echo_data(get_line(created_merge_details))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_error("There was an error creating merged VCF.")
            elif err.status_code == 404:
                self.echo_error(f"Project {self.project_id} does not exist.")
            raise



"""Common utils used in multiple commands."""
import json
import os
import uuid

import click

from gencove.logger import dump_debug_log, echo_error

map_arguments_to_human_readable = {
    "pipeline_capability_id": "Pipeline capability ID",
    "pipeline_id": "Pipeline ID",
    "project_id": "Project ID",
    "sample_ids": "sample IDs",
}


def handle_exception(message):
    """Exception handler for validation which raises an Abort"""
    echo_error(message)
    dump_debug_log()
    raise click.Abort()


def validate_uuid(ctx, param, candidate):  # pylint: disable=unused-argument
    """Test if provided string is a valid uuid version 4 string and convert
    to a hyphen uuid form if valid but no hyphens are present

    Note:
    Version 4 UUIDs have the form `xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx`
    where `x` is any hexadecimal digit and `y` is one of `8`, `9`, `a`, or `b`.

    candidate (str): uuid to check

    Returns:
        str: valid uuid v4, (8-4-4-4-12 form)

    Raises:
        Abort - if uuid is invalid.
    """
    try:
        # if version=4 attribute is set, value is silently converted
        validated_uuid = str(uuid.UUID(candidate))
    except ValueError:
        human_readable_param = map_arguments_to_human_readable.get(
            param.name, param.name
        )
        handle_exception(f"{human_readable_param} is not valid. Exiting.")
    return validated_uuid


def validate_uuid_list(ctx, param, uuids):  # pylint: disable=unused-argument
    """Test if provided sample_ids list contains only valid
    uuids when converted to a list.

    sample_ids (str): A comma separated list of uuids as a string, to check

    Returns:
        List[str]: a list of uuids

    Raises:
        Abort - if at least one uuid is invalid.
    """
    if uuids:
        try:
            uuids_list = [str(uuid.UUID(id.strip())) for id in uuids.split(",")]
        except ValueError:
            human_readable_param = map_arguments_to_human_readable.get(
                param.name, param.name
            )
            handle_exception(f"Not all {human_readable_param} are valid. Exiting.")
    return uuids_list if uuids else []


def sanitize_string(output):
    """Removes unwanted characters from output string."""
    return output.replace("\t", " ")


def is_valid_uuid(candidate):
    """Test if provided string is a valid uuid string.

    candidate (str): uuid to check

    Returns:
        bool: True if is a valid uuid, False if not
    """
    try:
        uuid.UUID(candidate)
        return True
    except ValueError:
        return False


def is_valid_json(candidate):
    """Test if provided string is a valid JSON.

    candidate (str): JSON to check

    Returns:
        bool: True if is a valid JSON, False if not
    """
    try:
        json.loads(candidate)
        return True
    except ValueError:
        return False


def validate_file_types(candidate_file_types, valid_file_types):
    """Test if provided list of file types is valid.

    candidate_file_types (tuple of str): File types to check
    valid_file_types (list of FileType): Valid file types to check against

    Returns:
        list: [] if candidate_file_types are valid, list of invalid file types if not
    """
    file_type_keys = [file_type.key for file_type in valid_file_types]

    # file types that are in candidate_file_types but not in valid_file_types
    return list(set(candidate_file_types).difference(set(file_type_keys)))


def validate_destination_exists(
    ctx, param, candidate
):  # pylint: disable=unused-argument
    """Validate that a destination is a directory and/or exists"""
    if not os.path.exists(candidate) or not os.path.isdir(candidate):
        human_readable_param = map_arguments_to_human_readable.get(
            param.name, param.name
        )
        handle_exception(
            f"{human_readable_param} is not a directory that exists. Exiting."
        )
    return candidate
